# -*- coding: utf-8 -*-
#
# Copyright © 2010 Nazar Vinnichuk <ux.isco@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ['VkQueue', 'VkQuery', 'VkException', 'VkError', 'VkSession']

import urllib, json, hashlib, urllib2, time, threading, subprocess

API_VERSION = '3.0'
API_URL     = 'http://api.vkontakte.ru/api.php'

import Queue

class VkQueue():
    def __init__(self, query, req_per_second = 5, threads_count = 5,
                 max_queue_size = 0):
        
        self.query = query
        
        self._queue = Queue.PriorityQueue(max_queue_size)
        
        req_interval = float(threads_count) / req_per_second
        for i in xrange(threads_count):
            VkThread(self._queue, self.query, req_interval).start()
    
    def add(self, method_name, method_params = {}, fok = None, ferror = None,
            prior = 0):
        
        event = VkEvent(fok, ferror)
        self._queue.put_nowait((prior, (method_name, method_params, event)))
        return event


class VkQuery:
    def __init__(self, session):
        self.session   = session
        self.app_id    = session.app_id
        self.test_mode = session.test_mode
    
    def do_request(self, method_name, method_params = {}):
        return self._do_request(method_name, method_params)
    
    def _do_request(self, method_name, method_params = {},
                    allow_update_session = True):
        
        session = self.session.get_session()
        
        params = dict(method_params, **{'api_id'    : str(self.app_id),
                                        'method'    : method_name,
                                        'v'         : API_VERSION,
                                        'format'    : 'JSON',
                                        'test_mode' : str(int(self.test_mode))})
        
        params['sig'] = self._make_sig(params, session)
        params['sid'] = session['sid']
        
        print 'params = %s' % params
        
        uri = API_URL + '?' + urllib.urlencode(params)
        
        print 'uri = %s' % uri
        
        response = json.load(urllib2.urlopen(uri))
        
        if 'error' in response:
            error = response['error']
            
            if allow_update_session and \
               (int(error['error_code']) == 3 or \
                int(error['error_code']) == 5):
                
                print 'vkontakte returned error: %s. updating session...' % \
                          error
                
                if not self.session.update_session():
                    raise VkException('Login failure')
                
                return self._do_request(method_name, method_params, False)
            
            else:
                raise VkError(error)
        
        elif 'response' in response:
            return response['response']
        
        else:
            raise VkException('Incorrect response')
    
    def _make_sig(self, params, session):
        sig_params = ['%s=%s' % (k, v) for k, v in params.items()]
        sig_params.sort()
        
        raw_sig = str(session['mid']) + ''.join(sig_params) + session['secret']
        
        print 'raw sig = %s' % raw_sig
        
        return hashlib.md5(raw_sig).hexdigest()


class VkException(Exception):
    def __init__(self, message):
        self.message = message
    
    def __str__(self):
        return 'VKontakte exception: ' % self.message


class VkError(Exception):
    def __init__(self, vk_error):
        self.code    = vk_error['error_code']
        self.message = vk_error['error_msg']
        self.params  = dict((param['key'], param['value']) \
                            for param in vk_error['request_params'])
    
    def __str__(self):
        return 'VKontakte error (%d): %s (Request params: %s)' % \
                   (self.code, self.message, self.params)


import os, glib

CONFIG_DIR = glib.get_user_config_dir() + '/pyvkapi'
PYTHON_PATH = '/usr/bin/python'
VK_AUTH_BROWSER_PATH = os.path.dirname(os.path.abspath(__file__)) + \
                       '/vk-auth-browser.py'

class VkSession:
    def __init__(self, app_id, settings = 0, test_mode = False,
                 session_updated_cb = None):
        
        self.app_id    = app_id
        self.settings  = settings
        self.test_mode = test_mode
        
        if (session_updated_cb is None or callable(session_updated_cb)):
            self._session_updated_cb = session_updated_cb
        else:
            raise Exception('Callback for session update is not callable')
        
        if not os.path.exists(CONFIG_DIR):
            os.mkdir(CONFIG_DIR, 0700)
        
        self._session_filename = CONFIG_DIR + '/%s-session.json' % self.app_id
        self._cookies_filename = CONFIG_DIR + '/%s-cookies.txt' % self.app_id
        
        self._session = None
    
    def get_session(self):
        if not self.has_session():
            try:
                print 'trying to read session from %s ...' % self._session_filename
                f = open(self._session_filename, 'r')
                self._session = json.loads(f.read())
                f.close()
                
                if self.has_session():
                    self._call_session_updated_cb()
                    print 'session = %s' % self._session
                else:
                    raise VkException('Empty session has been read')
            
            except:
                self.update_session()
        
        if not self.has_session():
            raise VkException('Error getting session. Maybe login failure.')
        
        return self._session
    
    def update_session(self):
        print 'updating session from server ...'
        
        args = [PYTHON_PATH, VK_AUTH_BROWSER_PATH, str(self.app_id),
                str(self.settings), self._cookies_filename]
        
        p = subprocess.Popen(stdout = subprocess.PIPE, args = args)
        stdout_strings = p.communicate()[0].split("\n")
        
        self._session = json.loads(stdout_strings[len(stdout_strings) - 2])
        
        for stdout_string in stdout_strings:
            print stdout_string
        
        print 'session = %s' % self._session
        
        self._call_session_updated_cb()
        self._store_session_items()
        
        return self.has_session()
    
    def delete_session(self):
        self._session = None
        
        os.remove(self._session_filename)
        os.remove(self._cookies_filename)
    
    def has_session(self):
        return self._session is not None
    
    def _call_session_updated_cb(self):
        if self._session_updated_cb is not None:
            self._session_updated_cb(self._session)
    
    def _store_session_items(self):
        print 'storing session to %s ...' % self._session_filename
        
        try:
            f = open(self._session_filename, 'w+')
            f.write(json.dumps(self._session))
            f.close()
        except:
            print 'errors occured, while trying to store session'
            return False
        else:
            print 'session stored'
            return True


FIND_LIBRARY_CMD = '/sbin/ldconfig -v 2>/dev/null | grep "%s" | cut -f3 -d" " | uniq | sort -V -r'

# ctypes.util.find_library works not realy good.
def find_library(name):
    p = subprocess.Popen(stdout = subprocess.PIPE, args = ['/bin/sh', '-c', FIND_LIBRARY_CMD % name])
    return p.communicate()[0].split("\n")[0]


BROWSER_TITLE  = 'ВКонтакте: вход'
BROWSER_WIDTH  = 500
BROWSER_HEIGHT = 325

LOGIN_URL = 'http://vkontakte.ru/login.php?layout=popup&type=browser&settings=%d&app=%d'
SUCCESS_URL_MATCH = '^http://vkontakte.ru/api/login_success.html#session=(.+)$'
FAILURE_URL_MATCH = '^http://vkontakte.ru/api/login_failure.html$'

import gtk, webkit, re, ctypes


class VkAuthDialog(gtk.Dialog):
    def __init__(self, app_id, settings = 0, cookies_filename = None):
        gtk.Dialog.__init__(self)
        
        self._uri = LOGIN_URL % (settings, app_id)
        
        if cookies_filename is not None:
            try:
                libsoup   = ctypes.CDLL(find_library('soup'))
                libwebkit = ctypes.CDLL(find_library('webkit'))
            except:
                raise Exception('Could not load either soup or webkit library')
            else:
                session = libwebkit.webkit_get_default_session()
                libsoup.soup_session_add_feature(session, libsoup.soup_cookie_jar_text_new(cookies_filename, False))
        
        web_view = webkit.WebView()
        settings = web_view.get_settings()
        settings.set_property('enable-developer-extras', False)
        web_view.set_full_content_zoom(True)
        web_view.show_all()
        web_view.connect('navigation-requested', self._navigation_requested_cb)
        
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.props.hscrollbar_policy = gtk.POLICY_AUTOMATIC
        scrolled_window.props.vscrollbar_policy = gtk.POLICY_AUTOMATIC
        scrolled_window.add(web_view)
        scrolled_window.show_all()
        
        self.vbox.pack_start(scrolled_window)
        self.set_default_size(BROWSER_WIDTH, BROWSER_HEIGHT)
        self.set_position(gtk.WIN_POS_CENTER)
        self.set_title(BROWSER_TITLE)
        self.set_modal(True)
        
        self.web_view = web_view
    
    def get_session(self):
        return self.run()
    
    def run(self):
        print 'login uri = %s' % self._uri
        self._session = None
        self.web_view.load_uri(self._uri)
        gtk.Dialog.run(self)
        return self._session
    
    def _navigation_requested_cb(self, view, frame, networkRequest):
        uri = networkRequest.get_uri()
        
        if re.match(SUCCESS_URL_MATCH, uri):
            uri_session = re.findall(SUCCESS_URL_MATCH, uri)[0]
            self._session = json.loads(urllib.unquote(uri_session))
            self.response(1)
            return True
        
        elif re.match(FAILURE_URL_MATCH, uri):
            self._session = None
            self.response(0)
            return True
        
        return False


class VkEvent():
    def __init__(self, fok, ferror):
        self.event  = threading.Event()
        self.fok    = fok
        self.ferror = ferror
        self.e      = False
    
    def ok(self, data):
        self.data = data
        if (self.fok):
            self.fok(data)
        self.event.set()
    
    def error(self, exception):
        self.e = exception
        if self.ferror:
            self.ferror(self.e.code, self.e.message, self.e.params)
        self.event.set()
    
    def fatal(self, exception):
        self.e = exception
        self.event.set()
    
    def wait(self):
        self.event.wait()
        
        if self.e:
            raise self.e
        else:
            return self.data


DEFAULT_THREAD_SLEEP = 0.1

class VkThread(threading.Thread):
    def __init__(self, queue, query, req_interval):
        threading.Thread.__init__(self)
        self._query  = query
        self._queue  = queue
        self._req_interval = req_interval
    
    def run(self):
        while True:
            (prior, (method_name, method_params, event)) = self._queue.get()
            sleep = self._req_interval
            while True:
                t = time.time()
                try:
                    data = self._query.do_request(method_name, method_params)
                except VkError, e:
                    if e.code == 6:
                        time.sleep(sleep)
                        sleep += 0.5 * self._req_interval
                        continue
                    else:
                        event.error(e)
                        break
                except Exception, e:
                    event.fatal(e)
                    break
                else:
                    event.ok(data)
                    break
                finally:
                    t = self._req_interval + t - time.time()
                    if t > 0:
                        time.sleep(t)
                    else:
                        time.sleep(DEFAULT_THREAD_SLEEP)

